// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "RenderPipeline.hpp"
#include "game/DawnGame.hpp"
#include "display/mesh/QuadMesh.hpp"
#include "scene/SceneItem.hpp"

using namespace Dawn;

RenderPipeline::RenderPipeline(RenderManager *renderManager) {
  assertNotNull(renderManager);

  this->renderManager = renderManager;
}

void RenderPipeline::init() {

}

void RenderPipeline::render() {
  if(this->renderManager->game->scene != nullptr) {
    this->renderScene(this->renderManager->game->scene);
  }
}

void RenderPipeline::renderScene(Scene *scene) {
  assertNotNull(scene);

  // Render subscenes first.
  auto subSceneControllers = scene->findComponents<SubSceneController>();
  auto itSubScene = subSceneControllers.begin();
  while(itSubScene != subSceneControllers.end()) {
    auto subScene = (*itSubScene)->getSubScene();
    if(subScene == nullptr) {
      ++itSubScene;
      continue;
    }

    if((*itSubScene)->onlyUpdateUnpaused && scene->game->timeManager.isPaused) {
      ++itSubScene;
      continue;
    }

    this->renderScene(subScene);
    ++itSubScene;
  }

  // Now render backbuffers.
  auto backBuffer = this->renderManager->getBackBuffer();
  auto cameras = scene->findComponents<Camera>();
  Camera *backBufferCamera = nullptr;

  // First, render all non-backbuffer cameras.
  auto it = cameras.begin();
  while(it != cameras.end()) {
    RenderTarget *cameraTarget = (*it)->getRenderTarget();

    // Leave the backbuffer camera(s) to last, so we skip them. we do this so
    // that children framebuffers contain the CURRENT frame, not LAST frame.
    if(cameraTarget == backBuffer) {
      backBufferCamera = *it;
    } else {
      this->renderSceneCamera(scene, *it);
    }

    ++it;
  }

  // Now render the backbuffer camera.
  if(backBufferCamera == nullptr) return;
  this->renderSceneCamera(scene, backBufferCamera);
}

void RenderPipeline::renderSceneCamera(Scene *scene, Camera *camera) {
  assertNotNull(scene);
  assertNotNull(camera);
  auto meshes = scene->findComponents<MeshRenderer>();
  auto uiCanvasList = scene->findComponents<UICanvas>();
  auto renderTarget = camera->getRenderTarget();

  assertNotNull(renderTarget);

  renderTarget->bind();
  renderTarget->clear(
    RENDER_TARGET_CLEAR_FLAG_DEPTH |
    RENDER_TARGET_CLEAR_FLAG_COLOR
  );
  this->renderManager->setRenderFlags(
    RENDER_MANAGER_RENDER_FLAG_DEPTH_TEST |
    RENDER_MANAGER_RENDER_FLAG_BLEND
  );
  
  // Render all Meshes on the scene
  auto it = meshes.begin();
  while(it != meshes.end()) {
    auto mesh = *it;
    auto material = mesh->item->getComponent<Material>();

    // TODO: fallback material?
    if(material == nullptr) {
      ++it;
      continue;
    }

    auto shader = material->getShader();
    assertNotNull(shader);
    shader->bind();
    shader->setGlobalParameters(camera->projection, camera->transform->getWorldTransform());
    shader->setMeshParameters(mesh->item->transform.getWorldTransform());
    material->setShaderParameters();

    mesh->mesh->draw(MESH_DRAW_MODE_TRIANGLES, 0, -1);
    ++it;
  }


  // Now we only render world-absolute UI canvas'
  auto uiShader = this->renderManager->getUIShader();
  assertNotNull(uiShader);
  uiShader->bind();
  uiShader->setUICamera(camera->projection, camera->transform->getWorldTransform());
  this->renderManager->setRenderFlags(RENDER_MANAGER_RENDER_FLAG_DEPTH_TEST);

  auto it2 = uiCanvasList.begin();
  while(it2 != uiCanvasList.end()) {
    auto canvas = *it2;
  
    switch(canvas->drawType) {
      case UI_DRAW_TYPE_WORLD_ABSOLUTE:
        break;

      default:
        ++it2;
        continue;
    }
    
    auto it3 = canvas->children.begin();
    auto rootMatrix = canvas->transform->getWorldTransform();
    while(it3 != canvas->children.end()) {
      (*it3)->draw(uiShader,
        glm::translate(glm::scale(rootMatrix, glm::vec3(1.0f, -1.0f, 1.0f)), glm::vec3(0, -renderTarget->getHeight(), 0))
      );
      ++it3;
    }
    ++it2;
  }


  // Now render camera-relative UI
  this->renderManager->setRenderFlags(
    RENDER_MANAGER_RENDER_FLAG_BLEND
  );
  uiShader->setUICamera(
    glm::ortho(0.0f, renderTarget->getWidth(), renderTarget->getHeight(), 0.0f),
    glm::mat4(1.0f)
  );

  it2 = uiCanvasList.begin();
  while(it2 != uiCanvasList.end()) {
    auto canvas = *it2;

    switch(canvas->drawType) {
      case UI_DRAW_TYPE_WORLD_CAMERA_RELATIVE:
        break;

      default:
        ++it2;
        continue;
    }

    auto it3 = canvas->children.begin();
    auto rootMatrix = canvas->transform->getWorldTransform();
    while(it3 != canvas->children.end()) {
      (*it3)->draw(uiShader, rootMatrix);
      ++it3;
    }
    ++it2;
  }
}

RenderPipeline::~RenderPipeline() {

}