// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UIComponent.hpp"

using namespace Dawn;

UIComponent::UIComponent(UICanvas *canvas) {
  assertNotNull(canvas);
  this->canvas = canvas;
}

void UIComponent::updatePositions() {
  // X Alignment
  if(this->alignX == UI_COMPONENT_ALIGN_STRETCH) {
    if(this->parent == nullptr) {
      this->width = this->canvas->getWidth();
    } else {
      this->width = this->parent->getWidth();
    }
    this->relativeX = this->alignment[0];
    this->width -= (this->alignment[0] + this->alignment[2]);
  } else if(this->alignX == UI_COMPONENT_ALIGN_START) {
    this->relativeX = this->alignment[0];
    this->width = this->alignment[2];
  } else if(this->alignX == UI_COMPONENT_ALIGN_END) {
    this->width = this->alignment[0];
    if(this->parent == nullptr) {
      this->relativeX = this->canvas->getWidth();
    } else {
      this->relativeX = this->parent->getWidth();
    }
    this->relativeX -= this->width;
    this->relativeX -= this->alignment[2];
  } else if(this->alignX == UI_COMPONENT_ALIGN_MIDDLE) {
    this->width = this->alignment[2];
    if(this->parent == nullptr) {
      this->relativeX = this->canvas->getWidth();
    } else {
      this->relativeX = this->parent->getWidth();
    }
    this->relativeX = (this->relativeX / 2.0f) - (this->width / 2.0f) + this->alignment[0];
  }

  // Y Alignment
  if(this->alignY == UI_COMPONENT_ALIGN_STRETCH) {
    if(this->parent == nullptr) {
      this->height = this->canvas->getHeight();
    } else {
      this->height = this->parent->getHeight();
    }
    this->relativeY = this->alignment[1];
    this->height -= (this->alignment[1] + this->alignment[3]);
  } else if(this->alignY == UI_COMPONENT_ALIGN_START) {
    this->relativeY = this->alignment[1];
    this->height = this->alignment[3];
  } else if(this->alignY == UI_COMPONENT_ALIGN_END) {
    this->height = this->alignment[1];
    if(this->parent == nullptr) {
      this->relativeY = this->canvas->getHeight();
    } else {
      this->relativeY = this->parent->getHeight();
    }
    this->relativeY -= this->height;
    this->relativeY -= this->alignment[3];
  } else if(this->alignY == UI_COMPONENT_ALIGN_MIDDLE) {
    this->height = this->alignment[3];
    if(this->parent == nullptr) {
      this->relativeY = this->canvas->getHeight();
    } else {
      this->relativeY = this->parent->getHeight();
    }
    this->relativeY = (this->relativeY / 2.0f) - (this->height / 2.0f) + this->alignment[1];
  }

  this->relativeX = mathRound<float_t>(this->relativeX);
  this->relativeY = mathRound<float_t>(this->relativeY);
  this->width = mathRound<float_t>(this->width);
  this->height = mathRound<float_t>(this->height);

  // Update children
  auto it = this->children.begin();
  while(it != this->children.end()) {
    (*it)->updatePositions();
    ++it;
  }

  // Fire event
  eventAlignmentUpdated.invoke(
    this->width, this->height, this->relativeX, this->relativeY
  );
}

float_t UIComponent::getWidth() {
  return this->width;
}

float_t UIComponent::getHeight() {
  return this->height;
}

float_t UIComponent::getContentWidth() {
  return this->width;
}

float_t UIComponent::getContentHeight() {
  return this->height;
}

float_t UIComponent::getRelativeX() {
  return this->relativeX;
}

float_t UIComponent::getRelativeY() {
  return this->relativeY;
}

DawnGame * UIComponent::getGame() {
  return this->canvas->getGame();
}

Scene * UIComponent::getScene() {
  return this->canvas->getScene();
}

void UIComponent::setTransform(
  UIComponentAlign xAlign,
  UIComponentAlign yAlign,
  glm::vec4 alignment,
  float_t z
) {
  this->alignX = xAlign;
  this->alignY = yAlign;
  this->alignment = alignment;
  this->z = z;
  this->updatePositions();
}

void UIComponent::draw(UIShader *uiShader, glm::mat4 parentTransform) {
  // Calculate self transform matrix
  glm::mat4 selfTransform = parentTransform * glm::translate(
    glm::mat4(1.0f), glm::vec3(this->relativeX, this->relativeY, this->z)
  );

  // Draw Self
  this->drawSelf(uiShader, selfTransform);

  // Render children
  auto it = this->children.begin();
  while(it != this->children.end()) {
    (*it)->draw(uiShader, selfTransform);
    ++it;
  }
}

void UIComponent::addChild(UIComponent *child) {
  if(child->parent == this) return;
  if(child->parent != nullptr) child->parent->removeChild(child);
  this->children.push_back(child);
  child->parent = this;
  this->updatePositions();
}

void UIComponent::removeChild(UIComponent *child) {
  assertTrue(child->parent == this);
  auto it = this->children.begin();
  while(it != this->children.end()) {
    if(*it == child) {
      this->children.erase(it);
      break;
    }
    ++it;
  }
  child->parent = nullptr;
}

UIComponent::~UIComponent() {
  
}